from itertools import cycle
from copy import deepcopy
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import animation
from .colored_plot import colored_plot
import matplotlib.patheffects as path_effects
from tqdm import tqdm
import matplotlib.colors as mcolors
from collections.abc import Iterable
from collections import namedtuple
import stoked

patches = namedtuple('patches', ['patches_type', 'args'])
def circle_patches(radius):
    return patches(plt.Circle, (radius,))

def ellipse_patches(rx, ry):
    return patches(mpl.patches.Ellipse, (2*rx, 2*ry))

def atleast(array, dim, length, dtype=None):
    """Given an n-dimensional array, return either an n or n+1 dimensional repeated array
        array      input array (or scalar)
        dim        dimension of the returned array (must be n or n+1)
        dtype      array datatype (default None)
    """

    ret = np.asarray(np.atleast_1d(array), dtype=dtype)
    if (dim not in (ret.ndim,ret.ndim+1)):
        raise ValueError('dim = {0} is invalid with input dim = {1}'.format(dim, ret.ndim))

    if len(ret.shape) == dim and ret.shape[0] != length:
        ret = np.repeat(ret, length)
    elif len(ret.shape) != dim:
        ret = np.array([ret]*length)

    return ret

def rotation_transform(axis, angle, ax = None):
    """Return a rotation transfrom that rotates around an axis by an angle
            axix[2]       (x,y) center of rotation
            angle         angle of rotation (in degrees)  """

    if ax is None: ax = plt.gca()
    t_scale = ax.transData
    t_rotate = mpl.transforms.Affine2D().rotate_deg_around(axis[0], axis[1], angle*180/np.pi)
    return t_rotate + t_scale

#TODO use matplotlib collections, patchcollection instead of python lists for performance
def animation_2d(func, patches, frames=None, colors=None, ax=None, time=None, time_unit='T',
        xlim=None, ylim=None, number_labels=False, trail=100, trail_type='fading',
        time_kwargs={}, label_kwargs={}, circle_kwargs={}, trail_kwargs={}, fading_kwargs={}, **kwargs):
    """Create a 2D animation of trajectories

       Arguments:
            func                       function that returns (x,y) coordinates and angles
            colors                     list of colors to cycle through
            ax (default None)          specify the axes of the animation
            time[N]                    display the time (in time_units)
            xlim[2]                    min,max values of x-axis
            ylim[2]                    min,max values of y-axis
            time_unit                  string label for the units of time (default 'T')
            number_labels              include text labels (1,2,...) per particle
            trail                      length of particle trail
            trail_type                 'normal' or 'fading'
            time_kwargs                additional arguments to pass to timer text object
            label_kwargs               additional arguments to pass to label text objects
            circle_kwargs              additional arguments to circle properites
            trail_kwargs               additional arguments to line trail properies
            fading_kwargs              Fading line properites, {max_lw, min_lw}
            kwargs                     Additional kwargs for FuncAnimation
    """
    trail_types = ['normal', 'fading']
    if trail_type not in trail_types:
        raise ValueError("trail_type '{}' is not valid. Choose from {}".format(trail_type, trail_types))
    if (trail_type == 'fading' and trail == np.inf):
        raise ValueError("trail cannot be fading and infinite")

    if colors is None:
        colors = mcolors.TABLEAU_COLORS
    if isinstance(colors, str):
        colors = [colors]
    if not isinstance(colors, Iterable):
        colors = [colors]

    time_properties = dict(fontsize=12, zorder=2)
    time_properties.update(time_kwargs)

    label_properties = dict(horizontalalignment='center', verticalalignment='center', fontsize=14, zorder=2)
    label_properties.update(label_kwargs)

    trail_properties = dict(zorder=0)
    trail_properties.update(trail_kwargs)

    circle_properties = dict(linewidth=1, zorder=1)
    circle_properties.update(circle_kwargs)

    line_properties = deepcopy(circle_properties)
    # line_properties.pop('facecolor')

    fading_properties = dict(max_lw=2, min_lw=0.3)
    fading_properties.update(fading_kwargs)

    positions, angles = func(0)
    positions = np.asarray(positions, dtype=float)
    if angles is not None:
        angles = np.asarray(angles, dtype=float)
    Nparticles = positions.shape[0]

    if ax is None:
        ax = plt.gca()
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    color_cycle = cycle(colors)

    particles = []
    dots = []
    lines = []
    trails = []
    text = {}

    if patches is not None:
        if not isinstance(patches.patches_type, Iterable):
            patches_type = [patches.patches_type]*Nparticles
        else:
            patches_type = patches.patches_type

        if not isinstance(patches.args[0], Iterable):
            patches_args = [patches.args for i in range(Nparticles)]
        else:
            patches_args = [[arg[i] for arg in patches.args] for i in range(Nparticles)]

    inv = ax.transData.transform
    for i in range(Nparticles): 
        pos = positions[i]
        color = next(color_cycle)

        if patches is not None:
            particles.append(patches_type[i](pos, *patches_args[i], fc=color, edgecolor='k', animated=False, **circle_properties))
            ax.add_patch(particles[-1])

            if patches_type[i] is plt.Circle and angles is not None:
                radius = patches_args[i][0]
                lines.append(plt.Line2D([pos[0]-radius, pos[0]+radius], [pos[1], pos[1]], lw=circle_properties['linewidth'], color=color, animated=False, **line_properties))
                ax.add_line(lines[-1])
        else:
            dots.append(plt.Circle(inv(pos[:2]), 3, color=color, animated=False, transform=None))
            ax.add_patch(dots[-1])

        if trail > 0:
            if trail_type == 'normal':
                trails.append(ax.plot([pos[0]], [pos[1]], color=color, **trail_properties)[0])
            elif trail_type == 'fading':
                c = np.zeros((trail,4))
                c[:,0:3] = mpl.colors.to_rgb(color)
                c[:,3] = np.linspace(1,0,trail)
                lw = np.linspace(fading_properties['max_lw'],fading_properties['min_lw'],trail)
                trails.append(colored_plot([pos[0]], [pos[1]], c, ax=ax, linewidth=lw, **trail_properties))

        
        if number_labels:
            label = str(i+1)
            text[label] = ax.text(pos[0], pos[1], label, animated=True, **label_properties)
            text[label].set_path_effects([path_effects.Stroke(linewidth=5, foreground='white'),
                       path_effects.Normal()])

    ax.set_aspect('equal')

    if time is not None:
        text['clock'] = ax.text(.98,0.02, r"{0:.2f} {1}".format(0.0, time_unit), transform=ax.transAxes, horizontalalignment='right', animated=True, **time_properties)

    # history = np.zeros((trail,) + positions.shape, dtype=float)
    history = np.tile(positions, (trail,1,1))
    def update(t):
        nonlocal history

        positions, angles = func(t)
        positions = np.asarray(positions, dtype=float)
        if angles is not None:
            angles = np.asarray(angles, dtype=float)

        history = np.roll(history, 1, axis=0)
        if trail > 0:
            history[0] = positions

        for i in range(Nparticles): 
            pos = positions[i]
            if particles:
                particles[i].center = pos
                if angles is not None:
                    tran = mpl.transforms.Affine2D().rotate_around(*pos[:2], angles[i])
                    particles[i].set_transform(tran + ax.transData)

                if angles is not None and patches_type[i] is plt.Circle:
                    radius = patches_args[i][0]
                    lines[i].set_data([pos[0]-radius, pos[0]+radius], [pos[1], pos[1]])
                    lines[i].set_transform(rotation_transform(pos[:2], angles[i], ax=ax))

            if dots:
                dots[i].center = inv(pos[:2])

            if time is not None:
                text['clock'].set_text(r"{0:.2f} {1}".format(time[t], time_unit))

            if trail > 0:
                trails[i].set_data(history[:,i,0], history[:,i,1])
            
            if number_labels:
                text[str(i+1)].set_position(pos)
                
        
        return  trails + particles + lines + list(text.values()) + dots

    def init():
        positions, angles = func(0)
        positions = np.asarray(positions, dtype=float)
        angles = np.asarray(angles, dtype=float)

        history[...] = np.tile(positions, (trail,1,1))
        return  trails + particles + lines + list(text.values()) + dots

    kw = dict(interval=30, repeat=True, blit=True)
    kw.update(kwargs)
    anim = animation.FuncAnimation(ax.figure, update, frames=frames, init_func=init, **kw)
    return anim

def trajectory_animation(trajectory, patches=None, *args, **kwargs):
    """
    Create a 2D animation give trajectory data

    Arguments:
        trajectory     trajectory data for T steps, N particles
        patches        optional patches object to represent particles
    """
    if not isinstance(trajectory, stoked.trajectory):
        trajectory = stoked.trajectory(trajectory)

    if trajectory.orientation is not None:
        angles = stoked.quaternion_to_angles(trajectory.orientation)
    else:
        angles = [None]*len(trajectory.position)

    def func(i):
        return trajectory.position[i], angles[i]

    xlim = np.array([np.min(trajectory.position[...,0]),
                     np.max(trajectory.position[...,0])])
    ylim = np.array([np.min(trajectory.position[...,1]),
                     np.max(trajectory.position[...,1])])
    if patches:
        xbuff = np.max(patches.args)
        ybuff = np.max(patches.args)
        xlim[0] -= xbuff
        xlim[1] += xbuff
        ylim[0] -= ybuff
        ylim[1] += ybuff
        


    xbuff = np.abs(xlim[1] - xlim[0])*0.1
    xlim += (-xbuff, xbuff)
    ybuff = np.abs(ylim[1] - ylim[0])*0.1
    ylim += (-ybuff, ybuff)

    return animation_2d(func, patches=patches, frames=len(trajectory.position), xlim=xlim, ylim=ylim, *args, **kwargs)



import numpy as np
from abc import ABCMeta, abstractmethod
from scipy.integrate import quad

class drag:
    """
    Abstract base class for drag coeffecients
    """
    __metaclass__ = ABCMeta

    def __init__(self, viscosity, isotropic=False):
        """
        Arguments:
            viscosity      dynamic viscosity µ of surrounding fluid
            isotropic      (bool) True if the drag is isotropic
        """
        self.viscosity = viscosity
        self.isotropic = isotropic

    @abstractmethod
    def _drag_T(self):
        raise NotImplementedError('translational drag has not be implemented for this type of particle')

    @abstractmethod
    def _drag_R(self):
        raise NotImplementedError('rotational drag has not be implemented for this type of particle')

    @property
    def drag_T(self):
        """translational drag coeffecient"""
        return self._drag_T()

    @property
    def drag_R(self):
        """rotational drag coeffecient"""
        return self._drag_R()

class drag_sphere(drag):
    """
    Drag coeffecients for a sphere
    """

    def __init__(self, radius, viscosity):
        """
        Arguments:
            radius[N]      sphere radii
            viscosity      dynamic viscosity µ of surrounding fluid
        """
        super().__init__(viscosity, isotropic=True)
        self.radius = np.asarray(radius, dtype=float)

    def _drag_T(self):
        return 6*np.pi*self.radius*self.viscosity

    def _drag_R(self):
        return 8*np.pi*self.radius**3*self.viscosity

class drag_ellipsoid(drag):
    """
    Drag coeffecients for an ellipsoid
    """

    def __init__(self, radii, viscosity):
        """
        Arguments:
            radii[N,3]     ellipsoid radii
            viscosity      dynamic viscosity µ of surrounding fluid
        """
        super().__init__(viscosity)
        self.radii = np.atleast_2d(np.asarray(radii, dtype=float))
        self.chi_0 = np.zeros(len(self.radii), dtype=float)
        self.alpha_0 = np.zeros_like(self.radii)

        for i in range(len(self.radii)):
            a, b, c = self.radii[i]
            integrand = lambda t: 1/np.sqrt((1 + t)*((b/a)**2 + t)*((c/a)**2 + t))
            self.chi_0[i] = b*c*quad(integrand, 0, np.inf)[0]

            for j, comp in enumerate([a,b,c]):
                integrand = lambda t: 1/((1 + t)*np.sqrt(((a/comp)**2 + t)*((b/comp)**2 + t)*((c/comp)**2 + t)))
                self.alpha_0[i][j] = a*b*c/comp*quad(integrand, 0, np.inf)[0]

    def _drag_T(self):
        D = np.zeros_like(self.radii)
        for i in range(len(self.radii)):
            a, b, c = self.radii[i]
            for j, comp in enumerate([a,b,c]):
                D[i][j] = 16*np.pi*self.viscosity*a*b*c/(self.chi_0[i] + self.alpha_0[i][j])

        return D

    def _drag_R(self):
        D = np.zeros_like(self.radii)
        for i in range(len(self.radii)):
            a, b, c = self.radii[i]
            factor = 16*np.pi*self.viscosity*a*b*c 

            D[i][0] = factor/(3*(self.alpha_0[i][1] + self.alpha_0[i][2]))*(b**2 + c**2)
            D[i][1] = factor/(3*(self.alpha_0[i][0] + self.alpha_0[i][2]))*(a**2 + c**2)
            D[i][2] = factor/(3*(self.alpha_0[i][0] + self.alpha_0[i][1]))*(a**2 + b**2)

        return D



import numpy as np

def levi_civita():
    """return the levi-civita symbol"""

    eijk = np.zeros((3, 3, 3), dtype=float)
    eijk[0, 1, 2] = eijk[1, 2, 0] = eijk[2, 0, 1] = 1
    eijk[0, 2, 1] = eijk[2, 1, 0] = eijk[1, 0, 2] = -1
    return eijk

def grand_mobility_matrix(position, drag_T, drag_R, viscosity):
    """
    Construct the grand mobility matrix for a given cluster

    Arguments:
        position[N,3]     position of N particles
        drag_T[N,3,3]     3 by 3 translational drag tensors of N particles
        drag_R[N,3,3]     3 by 3 rotational drag tensors of N particles
        viscosity         dynamic viscosity µ of surrounding fluid
    """

    Nparticles = len(position)
    M = np.zeros([2, 3*Nparticles, 2, 3*Nparticles], dtype=float)

    ### block-diagonal components
    for i in range(Nparticles):
        idx = np.s_[0,3*i:3*i+3,0,3*i:3*i+3]
        M[idx] = drag_T[i]

        idx = np.s_[1,3*i:3*i+3,1,3*i:3*i+3]
        M[idx] = drag_R[i]

    ### Off block-diagonal components
    factor = 1/(8*np.pi*viscosity)
    eps = levi_civita()
    for i in range(Nparticles):
        for j in range(i+1, Nparticles):
            r_ijx = position[i] - position[j]
            r_ij = np.linalg.norm(r_ijx)

            I = np.identity(3, dtype=float)
            T = np.outer(r_ijx, r_ijx)/r_ij**2
            K = np.einsum('ijk,k->ij', eps, r_ijx)/r_ij

            ### TT coupling
            idx = np.s_[0,3*i:3*i+3,0,3*j:3*j+3]
            M[idx] = factor/r_ij*(I + T)
            idx2 = np.s_[0,3*j:3*j+3,0,3*i:3*i+3]
            M[idx2] = M[idx]

            ### RR coupling
            idx = np.s_[1,3*i:3*i+3,1,3*j:3*j+3]
            M[idx] = factor/(2*r_ij**3)*(3*T - I)
            idx2 = np.s_[1,3*j:3*j+3,1,3*i:3*i+3]
            M[idx2] = M[idx]

            ### RT coupling
            idx = np.s_[1,3*i:3*i+3,0,3*j:3*j+3]
            M[idx] = -factor/r_ij**2*(K)
            idx2 = np.s_[1,3*j:3*j+3,0,3*i:3*i+3]
            M[idx2] = -M[idx]

            ### TR coupling
            idx3 = np.s_[0,3*i:3*i+3,1,3*j:3*j+3]
            M[idx3] = -M[idx]
            idx4 = np.s_[0,3*j:3*j+3,1,3*i:3*i+3]
            M[idx4] = -M[idx2]

    return M.reshape([6*Nparticles, 6*Nparticles])



import numpy as np
from stoked import interactions

class pairwise_force(interactions):
    """
    Pair-wise interaction force
    """
    def __init__(self, force_func):
        """
        Arguments:
            force_func     force function of the form F(r[dim]) -> [dim]
        """
        self.force_func = force_func

    def force(self):
        Nparticles = len(self.position)

        r_ij = self.position[:,np.newaxis] - self.position[np.newaxis]   # N x N x 3
        with np.errstate(divide='ignore'):
            F_ij = self.force_func(r_ij)   # N x N x 3
        np.einsum('iix->ix', F_ij)[...] = 0
        F_i = np.sum(F_ij, axis=1)

        return F_i

    def torque(self):
        T_i = np.zeros_like(self.position)
        return T_i

    def __add__(self, other):
        def new_func(r):
            return self.force_func(r) + other.force_func(r)

        return pairwise_force(new_func)

class pairwise_central_force(interactions):
    def __init__(self, force_func):
        """
        Arguments:
            force_func     force function of the form F(r[dim]) -> [dim]
        """
        self.force_func = force_func

    def force(self):
        Nparticles = len(self.position)

        r_ij = self.position[:,np.newaxis] - self.position[np.newaxis]   # N x N x 3
        r = np.linalg.norm(r_ij, axis=-1)
        with np.errstate(divide='ignore'):
            F = self.force_func(r)
            r_inv = 1/r

        F_ij = np.einsum('ij,ijx,ij->ijx', F, r_ij, r_inv)
        np.einsum('iix->ix', F_ij)[...] = 0
        F_i = np.sum(F_ij, axis=1)

        return F_i

    def torque(self):
        T_i = np.zeros_like(self.position)
        return T_i

    def __add__(self, other):
        def new_func(r):
            return self.force_func(r) + other.force_func(r)

        return pairwise_central_force(new_func)

def pairwise_potential(potential_func):
    pass
    # def F(rvec):
        # eps = 1e-15*np.ones_like(rvec)
        # U1 = potential_func(rvec)
        # U2 = potential_func(rvec + eps)
         # = -(U2 - U1)/eps


    # return pairwise_force(F)

def pairwise_central_potential():
    pass



from scipy import constants
import numpy as np

density_water = 997

def gravity(volume, density, density_medium=density_water, direction=[0,0,-1]):
    direction = np.asarray(direction, dtype=float)

    def force(time, position, orientation):
        if np.isscalar(volume):
            vol = np.full(len(position), volume)
        else:
            vol = np.asarray(volume, dtype=float)

        if np.isscalar(density):
            den = np.full(len(position), density)
        else:
            den = np.asarray(density, dtype=float)

        mass = vol*(den - density_medium)
        F = constants.g*mass
        return np.outer(F, direction)

    return force

def gravity_sphere(radius, density, density_medium=density_water, direction=[0,0,-1]):
    volume = 4/3*np.pi*radius**3
    return gravity(volume, density, density_medium, direction)

def gravity_ellipsoid(radii, density, density_medium=density_water, direction=[0,0,-1]):
    volume = 4/3*np.pi*np.product(radii, axis=-1)
    return gravity(volume, density, density_medium, direction)



import numpy as np
from scipy.constants import k as kb
from collections.abc import Iterable
import quaternion
from abc import ABCMeta, abstractmethod
from .hydrodynamics import grand_mobility_matrix
from tqdm import tqdm

class trajectory:
    def __init__(self, position, orientation=None):
        self.position = np.asarray(position, dtype=float)

        if orientation is not None:
            self.orientation = np.asarray(orientation, dtype=np.quaternion)
        else:
            self.orientation = None

class interactions:
    """
    Abstract base class for particle interactions
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def force(self):
        """
        Return the force[N,3] based on the current state
        """
        pass

    @abstractmethod
    def torque(self):
        """
        Return the torque[N,3] based on the current state
        """
        pass
    
    @abstractmethod
    def update(self):
        """
        Update the current state
        """
        pass

    def _update(self, time, position, orientation, temperature):
        self.time = time
        self.position = position
        self.orientation = orientation
        self.temperature = temperature
        self.update()

def fluctuation(friction, T, dt):
    """Given a friction matrix, return the fluctuation matrix"""
    if T == 0:
        return np.zeros_like(friction)
    else:
        rhs = 2*kb*T*friction/dt
        return np.linalg.cholesky(rhs)

class stokesian_dynamics:
    """
    Perform a Stokesian dynamics simulation, with optional external and internal internal interactions and rotational dynamics
    """
    def __init__(self, *, temperature, dt, position, drag, orientation=None, 
                 force=None, torque=None, interactions=None, constraint=None,
                 hydrodynamic_coupling=True):
        """
        Arguments:
            temperature        system temperature
            dt                 time-step
            position[N,D]      initial position of N particles in D dimensions
            drag               drag coefficients (of base type stoked.drag)
            orientation[N]     initial orientation of N particles (as quaternions)
            force(t, r, q)     external force function given time t, position r[N,D], orientation q[N] and returns force F[N,D] (can be a list of functions)
            torque(t, r, q)    external torque function given time t, position r[N,D], orientation q[N] and returns torque T[N,D] (can be a list of functions)
            interactions       particle interactions (can be a list)
            constraint         particle constraints (can be a list)
            hydrodynamic_coupling    if True, include hydrodynamic coupling interactions (default: True)
        """
        self.temperature = temperature
        self.dt = dt
        self.time = 0
        self.position = np.atleast_2d(np.asarray(position, dtype=float))
        self.drag = drag
        self.Nparticles = len(self.position)
        self.ndim = self.position.shape[1]
        self.hydrodynamic_coupling = hydrodynamic_coupling

        if orientation is not None:
            self.orientation = np.atleast_1d(np.asarray(orientation, dtype=np.quaternion))
        else:
            self.orientation = np.ones([self.Nparticles], dtype=np.quaternion)

        if force is None:
            self.force = [lambda t, rvec, orient: np.zeros_like(rvec)]
        elif not isinstance(force, Iterable):
            self.force = [force]
        else:
            self.force = force

        if torque is None:
            self.torque = [lambda t, rvec, orient: np.zeros_like(rvec)]
        elif not isinstance(force, Iterable):
            self.torque = [torque]
        else:
            self.torque = torque

        if interactions is None:
            self.interactions = None
        elif not isinstance(interactions, Iterable):
            self.interactions = [interactions]
        else:
            self.interactions = interactions
        self._combine_pairwise_interactions()

        if constraint is None:
            self.constraint = []
        elif not isinstance(constraint, Iterable):
            self.constraint = [constraint]
        else:
            self.constraint = constraint

        self.alpha_T = np.zeros_like(self.position)
        drag_T = self.drag.drag_T
        if self.drag.isotropic and not np.isscalar(drag_T):
            drag_T = drag_T[:,np.newaxis]

        self.alpha_T[...] = 1/drag_T
        self.velocity = np.zeros_like(self.position)

        if self.orientation is not None:
            self.alpha_R = np.zeros_like(self.position)
            drag_R = self.drag.drag_R
            if self.drag.isotropic and not np.isscalar(drag_R):
                drag_R = drag_R[:,np.newaxis]

            self.alpha_R[...] = 1/drag_R
            self.angular_velocity = np.zeros_like(self.position)

        self.isotropic = self.drag.isotropic and not self.hydrodynamic_coupling
        if self.isotropic and orientation is None:
            self.rotating = False
        else:
            self.rotating = True

    def step(self):
        """
        Time-step the positions (and orientations) by dt
        """
        if self.hydrodynamic_coupling:
            self._step_hydrodynamics()
        else:
            self._step()

    def run(self, Nsteps, progress=True):
        """
        Run the simulation for Nsteps, returning the trajectories

        Arguments:
            Nsteps       number of steps to run
            progress     If True, display a progress bar
        """
        position = np.zeros((Nsteps,) + self.position.shape, dtype=float)

        if self.rotating:
            orientation = np.zeros((Nsteps,) + self.orientation.shape, dtype=np.quaternion)
        else:
            orientation = None

        for i in tqdm(range(Nsteps), desc='Running dynamics', disable=(not progress)):
            position[i] = self.position
            if self.rotating:
                orientation[i] = self.orientation
            self.step()

        return trajectory(position, orientation)

    def run_until(self, condition):
        """
        Run the simulation until some condition is met, returning the trajectories

        Arguments:
            condition    Function to return True when simulation should terminate
            progress     If True, display a progress bar
        """
        position = []

        if self.rotating:
            orientation = []
        else:
            orientation = None

        with tqdm(desc='Running dynamics until condition is met', disable=(not progress)) as pbar:
            while not condition():
                position.append(np.copy(self.position))
                if self.rotating:
                    orientation.append(np.copy(self.orientation))
                self.step()
                pbar.update()

        position = np.array(position, dtype=float)
        if self.rotating:
            orientation = np.array(orientation, dtype=np.quaternion)

        return trajectory(position, orientation)

    def _step(self):
        self._update_interactions(self.time, self.position, self.orientation)

        F = self._total_force(self.time, self.position, self.orientation)
        noise_T = np.random.normal(size=self.position.shape) 
        v1 = self._get_velocity(self.alpha_T, F, noise_T, self.orientation)
        r_predict = self.position + self.dt*v1

        if self.rotating:
            T = self._total_torque(self.time, self.position, self.orientation)
            noise_R = np.random.normal(size=self.position.shape) 
            w1 = self._get_velocity(self.alpha_R, T, noise_R, self.orientation)
            w1_q = np.array([np.quaternion(*omega) for omega in w1])
            o_predict = (1 + w1_q*self.dt/2)*self.orientation
        else:
            o_predict = self.orientation

        self.time += self.dt
        self._perform_constraints(r_predict, o_predict)
        self._update_interactions(self.time, r_predict, o_predict)

        F_predict = self._total_force(self.time, r_predict, o_predict)
        v2 = self._get_velocity(self.alpha_T, F_predict, noise_T, o_predict)
        self.velocity = 0.5*(v1 + v2)
        self.position += self.dt*self.velocity

        if self.rotating:
            T_predict =  self._total_torque(self.time, r_predict, o_predict)
            w2 = self._get_velocity(self.alpha_R, T_predict, noise_R, o_predict)
            w2_q = np.array([np.quaternion(*omega) for omega in w2])
            w_q = (w1_q + w2_q)/2
            self.angular_velocity = 0.5*(w1 + w2)
            self.orientation = (1 + w_q*self.dt/2)*self.orientation

        self._perform_constraints(self.position, self.orientation)
        if self.rotating:
            self.orientation = np.normalized(self.orientation)

    def _perform_constraints(self, position, orientation):
        for constraint in self.constraint:
            constraint(position, orientation)

    def _step_hydrodynamics(self):
        self._update_interactions(self.time, self.position, self.orientation)

        F = self._total_force(self.time, self.position, self.orientation)
        noise_T = np.random.normal(size=self.position.shape) 

        T = self._total_torque(self.time, self.position, self.orientation)
        noise_R = np.random.normal(size=self.position.shape) 

        v1, w1 = self._get_velocity_hydrodynamics(F, T, noise_T, noise_R, self.orientation)

        r_predict = self.position + self.dt*v1
        w1_q = np.array([np.quaternion(*omega) for omega in w1])
        o_predict = (1 + w1_q*self.dt/2)*self.orientation


        self.time += self.dt
        self._perform_constraints(r_predict, o_predict)
        self._update_interactions(self.time, r_predict, o_predict)

        F_predict = self._total_force(self.time, r_predict, o_predict)
        T_predict =  self._total_torque(self.time, r_predict, o_predict)

        v2, w2 = self._get_velocity_hydrodynamics(F_predict, T_predict, noise_T, noise_R, o_predict)

        w2_q = np.array([np.quaternion(*omega) for omega in w2])
        w_q = (w1_q + w2_q)/2

        self.velocity = 0.5*(v1 + v2)
        self.position += self.dt*self.velocity
        self.angular_velocity = 0.5*(w1 + w2)
        self.orientation = (1 + w_q*self.dt/2)*self.orientation

        self._perform_constraints(self.position, self.orientation)
        self.orientation = np.normalized(self.orientation)

    def _get_velocity(self, alpha, drive, noise, orientation):
        """
        FOR INTERNAL USE ONLY

        obtain velocity (or angular velocity) given alpha parameter, drive force/torque, noise, and orientation
        """
        if self.isotropic:
            beta = np.sqrt(2*kb*self.temperature*alpha/self.dt)
            velocity = alpha*drive + beta*noise
        else:
            rot = quaternion.as_rotation_matrix(orientation)
            alpha = np.einsum('Nij,Nj,Nlj->Nil', rot, alpha, rot)

            beta = np.zeros_like(alpha)
            for i in range(self.Nparticles):
                beta[i] = fluctuation(alpha[i], self.temperature, self.dt)

            velocity = np.einsum('Nij,Nj->Ni', alpha, drive) + np.einsum('Nij,Nj->Ni', beta, noise)

        return velocity

    def _get_velocity_hydrodynamics(self, F, T, noise_T, noise_R, orientation):
        """
        FOR INTERNAL USE ONLY
        """
        drag_T = np.zeros([self.Nparticles, 3, 3], dtype=float)
        drag_R = np.zeros([self.Nparticles, 3, 3], dtype=float)
        if self.drag.isotropic:
            np.einsum('Nii->Ni', drag_T)[...] = self.alpha_T
            np.einsum('Nii->Ni', drag_R)[...] = self.alpha_R
        else:
            rot = quaternion.as_rotation_matrix(orientation)
            drag_T[...] = np.einsum('Nij,Nj,Nlj->Nil', rot, self.alpha_T, rot)
            drag_R[...] = np.einsum('Nij,Nj,Nlj->Nil', rot, self.alpha_R, rot)

        M = grand_mobility_matrix(self.position, drag_T, drag_R, self.drag.viscosity)

        grand_F = np.hstack([F.flatten(), T.flatten()]) 
        grand_noise = np.hstack([noise_T.flatten(), noise_R.flatten()]) 
        N = fluctuation(M, self.temperature, self.dt)

        grand_velocity = M @ grand_F + N @ grand_noise
        velocity = grand_velocity[:3*self.Nparticles].reshape([self.Nparticles,3])
        angular_velocity = grand_velocity[3*self.Nparticles:].reshape([self.Nparticles,3])

        return velocity, angular_velocity

    def _update_interactions(self, time, position, orientation):
        """
        FOR INTERNAL USE ONLY

        update the interactions given a new (or predicted) position and orientation
        """
        if self.interactions is not None:
            for I in self.interactions:
                I._update(time, position, orientation, self.temperature)

    def _total_force(self, time, position, orientation):
        """
        FOR INTERNAL USE ONLY

        return the total force at a given time, position, and orientation
        """
        F = sum((force(time, position, orientation) for force in self.force))
        if self.interactions is not None:
            for I in self.interactions:
                F += I.force()
        return F

    def _total_torque(self, time, position, orientation):
        """
        FOR INTERNAL USE ONLY

        return the total torque at a given time, position, and orientation
        """
        T = sum((torque(time, position, orientation) for torque in self.torque))
        if self.interactions is not None:
            for I in self.interactions:
                T += I.torque()
        return T

    def _combine_pairwise_interactions(self):
        from stoked.forces import pairwise_central_force

        if self.interactions is not None:
            pairwise_forces = []
            for i,I in enumerate(self.interactions):
                if isinstance(I, pairwise_central_force):
                    pairwise_forces.append(I)

            if len(pairwise_forces) > 1:
                for pf in pairwise_forces:
                    self.interactions.remove(pf)
                self.interactions.append(sum(pairwise_forces[1:], start=pairwise_forces[0]))

def brownian_dynamics(*, temperature, dt, position, drag, orientation=None, 
                 force=None, torque=None, interactions=None, constraint=None):
    """
    Perform a Brownian dynamics simulation, with optional external and internal internal interactions and rotational dynamics
        
    Arguments:
        temperature        system temperature
        dt                 time-step
        position[N,D]      initial position of N particles in D dimensions
        drag               drag coefficients (of base type stoked.drag)
        orientation[N]     initial orientation of N particles (as quaternions)
        force(t, r, q)     external force function given time t, position r[N,D], orientation q[N] and returns force F[N,D] (can be a list of functions)
        torque(t, r, q)    external torque function given time t, position r[N,D], orientation q[N] and returns torque T[N,D] (can be a list of functions)
        interactions       particle interactions (can be a list)
        constraint         particle constraints (can be a list)
    """
    return stokesian_dynamics(temperature=temperature, dt=dt, position=position, drag=drag,
                  orientation=orientation, force=force, torque=torque, interactions=interactions,
                  constraint=constraint, hydrodynamic_coupling=False)
